# Copyright 2023 Nito T.M.
# License https://www.apache.org/licenses/LICENSE-2.0 Apache-2.0
# Author Nito T.M. (https://github.com/nitotm)
# Package pypi.org/project/eld/

import json
from eld.languageData import languageData


class LanguageResult:
    def __init__(self, results=None, num_ngrams=None):
        self.language = (languageData.lang_codes[results[0][0]] if results else None)
        self.__results = results
        self.__num_ngrams = num_ngrams

    def __str__(self):
        return json.dumps({'<object>': {
            'language': self.language,
            'scores()': self.scores(),
            'is_reliable()': self.is_reliable()
        }
        })

    def scores(self):
        return _get_scores(self.__results)

    def is_reliable(self):
        if not self.language or self.__num_ngrams < 3 or not self.__results:
            return False
        next_score = (self.__results[1][1] if len(self.__results) > 1 else 0)
        # A minimum of a 24% from the average score
        if languageData.avg_score[self.language] * 0.24 > (self.__results[0][1] / self.__num_ngrams) \
                or 0.01 > abs(self.__results[0][1] - next_score):
            return False
        return True


def _get_scores(results):
    scores = {}
    if results:
        for value in results:
            scores[languageData.lang_codes[value[0]]] = value[1]
    return scores



# Copyright 2023 Nito T.M.
# License https://www.apache.org/licenses/LICENSE-2.0 Apache-2.0
# Author Nito T.M. (https://github.com/nitotm)
# Package pypi.org/project/eld/

import importlib.util
import os


class LanguageData:
    def __init__(self):
        from .resources.avg_score import avg_score
        self.avg_score = avg_score
        self.ngrams = {}
        self.lang_score = []
        self.lang_codes = {}
        self.type = ''
        self.folder = os.path.dirname(__file__) + '/resources/ngrams/'

    # ISO 639-1 codes, for the 60 languages set.
    # ['am', 'ar', 'az', 'be', 'bg', 'bn', 'ca', 'cs', 'da', 'de', 'el', 'en', 'es', 'et', 'eu', 'fa', 'fi', 'fr', 'gu',
    # 'he', 'hi', 'hr', 'hu', 'hy', 'is', 'it', 'ja', 'ka', 'kn', 'ko', 'ku', 'lo', 'lt', 'lv', 'ml', 'mr', 'ms', 'nl',
    # 'no', 'or', 'pa', 'pl', 'pt', 'ro', 'ru', 'sk', 'sl', 'sq', 'sr', 'sv', 'ta', 'te', 'th', 'tl', 'tr', 'uk', 'ur',
    # 'vi', 'yo', 'zh']

    # ['Amharic', 'Arabic', 'Azerbaijani (Latin)', 'Belarusian', 'Bulgarian', 'Bengali', 'Catalan', 'Czech', 'Danish',
    # 'German', 'Greek', 'English', 'Spanish', 'Estonian', 'Basque', 'Persian', 'Finnish', 'French', 'Gujarati',
    # 'Hebrew', 'Hindi', 'Croatian', 'Hungarian', 'Armenian', 'Icelandic', 'Italian', 'Japanese', 'Georgian', 'Kannada',
    # 'Korean', 'Kurdish (Arabic)', 'Lao', 'Lithuanian', 'Latvian', 'Malayalam', 'Marathi', 'Malay (Latin)', 'Dutch',
    # 'Norwegian', 'Oriya', 'Punjabi', 'Polish', 'Portuguese', 'Romanian', 'Russian', 'Slovak', 'Slovene', 'Albanian',
    # 'Serbian (Cyrillic)', 'Swedish', 'Tamil', 'Telugu', 'Thai', 'Tagalog', 'Turkish', 'Ukrainian', 'Urdu',
    # 'Vietnamese', 'Yoruba', 'Chinese']

    def load_ngrams(self, subset_file=''):
        if subset_file == '':
            from .resources.ngrams.ngramsM60 import ngrams_data
        else:
            # module = importlib.import_module('.ngrams.' + subset_file)
            file_path = self.folder + subset_file + '.py'
            if not os.path.exists(file_path):
                file_path = self.folder + 'subset/' + subset_file + '.py'
            spec = importlib.util.spec_from_file_location(subset_file, file_path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            ngrams_data = module.ngrams_data

        self.ngrams = ngrams_data['ngrams']
        self.lang_score = [0] * (max(ngrams_data['languages'].keys()) + 1)
        self.type = ngrams_data['type']
        self.lang_codes = ngrams_data['languages']


languageData = LanguageData()



import pytest
import sys
import os

# Make sure, local package is imported instead of pip package
project_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))
sys.path.insert(0, project_root)  # prioritize the local package
# sys.path.append('../..')

from eld import LanguageDetector


# Mostly functional testing, when functions are more mature I will add some more unit tests

def test_load_eld():
    detector = LanguageDetector()
    assert isinstance(detector, LanguageDetector)


def test_dynamic_subset_detect():
    detector = LanguageDetector()
    lang_subset = ['en']
    detector.dynamic_lang_subset(lang_subset)
    result = len(detector.detect('How are you? Bien, gracias').scores())
    assert result == 1, 'Expected: 1 score, subset of only one language'


def test_remove_dynamic_subset():
    detector = LanguageDetector()
    lang_subset = ['en']
    detector.dynamic_lang_subset(lang_subset)
    detector.dynamic_lang_subset(None)
    result = len(detector.detect('How are you? Bien, gracias').scores())
    assert result > 1


def test_subset_detect():
    detector = LanguageDetector()
    lang_subset = ['en']
    detector.lang_subset(lang_subset)
    result = len(detector.detect('How are you? Bien, gracias').scores())
    assert result == 1, 'Expected: 1 score, subset of only one language'


def test_remove_subset():
    detector = LanguageDetector()
    lang_subset = ['en']
    detector.lang_subset(lang_subset)
    detector.lang_subset(None)
    result = len(detector.detect('How are you? Bien, gracias').scores())
    assert result > 1


def test_save_subset_file():
    # TODO use importlib or pathlib to check subset file as package resource
    file = os.path.dirname(__file__) + '/../resources/ngrams/subset/ngramsM60-1_2rrx014rx6ypsas6tplo1gtcnmiv5mz.py'
    if os.path.exists(file):
        os.remove(file)
    detector = LanguageDetector()
    lang_subset = ['en']
    detector.lang_subset(lang_subset)
    result = os.path.exists(file)
    assert result is True, 'Subset languages file Not saved: ' + file


def test_load_ngrams_detect():
    detector = LanguageDetector('ngramsM60-6_5ijqhj4oecs310zqtm8u9pgmd9ox2yd')
    result = detector.detect('Hola, cómo te llamas?').language
    assert result == 'es'



# Copyright 2023 Nito T.M.
# License https://www.apache.org/licenses/LICENSE-2.0 Apache-2.0
# Author Nito T.M. (https://github.com/nitotm)
# Package pypi.org/project/eld/

import hashlib
import os
import copy
import importlib.util
import logging

from eld.languageData import languageData
from eld.subsetResult import SubsetResult


class LanguageSubset:

    def __init__(self):
        self.subset = None
        self.default_ngrams = None
        self.loaded_subset = None

    def dynamic_lang_subset(self, languages):
        """
        Sets a subset, then detect() will filter the languages not included, from the scores with filterLangSubset()
        Call dynamic_lang_subset(None) to deactivate

        Args:
            languages (list or None): List of languages (ISO 639-1) to include in subset, or None to delete subset

        Returns:
            object SubsetResult: success (bool), languages (list or None), error (str or None)
        """
        self.subset = None
        if languages:
            self.subset = _make_subset(languages)
            if self.subset is None:
                return SubsetResult(False, None, 'No language matched this set')
        return SubsetResult(True, _iso_languages(self.subset) if self.subset else None)

    def lang_subset(self, languages, save=True):
        """
         Sets a subset and removes the excluded languages form the ngrams database
         if $save option is true, the new ngrams subset will be stored, and cached for next time

        Args:
            languages (list or None): List of languages (ISO 639-1) to include in subset, or None to delete subset

        Returns:
            object SubsetResult: success (bool), languages (list or None), error (str or None), file (str)
        """
        if not languages:
            if self.loaded_subset and self.default_ngrams:
                languageData.ngrams = copy.deepcopy(self.default_ngrams)
                self.loaded_subset = None
            return SubsetResult(True)  # if there was already no subset to disable, it also is successful

        lang_array = _make_subset(languages)
        if not lang_array:
            return SubsetResult(False, None, 'No language matched this set')

        if self.default_ngrams is None:
            self.default_ngrams = copy.deepcopy(languageData.ngrams)

        langs_str = [str(lang) for lang in lang_array]
        new_subset = base16_to_base36(
            hashlib.sha1(','.join(langs_str).encode()).hexdigest()
        )
        file_name = 'ngrams' + languageData.type + '-' + str(len(lang_array)) + '_' + new_subset
        file_path = languageData.folder + 'subset/' + file_name + '.py'

        if self.loaded_subset != new_subset:
            self.loaded_subset = new_subset

            if os.path.exists(file_path):
                # module = importlib.import_module('.ngrams.' + file_name, package=file_name)
                spec = importlib.util.spec_from_file_location(file_name, file_path)
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                languageData.ngrams = module.ngrams_data['ngrams']
                if languageData.ngrams:
                    return SubsetResult(True, _iso_languages(lang_array), None, file_path)

            if self.default_ngrams != languageData.ngrams:
                languageData.ngrams = copy.deepcopy(self.default_ngrams)

            for ngram, langsID in self.default_ngrams.items():
                for lid, value in langsID.items():
                    if lid not in lang_array:
                        del languageData.ngrams[ngram][lid]
                if not languageData.ngrams[ngram]:
                    del languageData.ngrams[ngram]

        saved = False
        if save:
            saved = _save_ngrams(file_path, lang_array)

        return SubsetResult(True, _iso_languages(lang_array), None, (file_name if saved else None))

    def _filter_lang_subset(self, scores):
        """Filters languages not included in the subset, from the result scores"""
        sub_results = []
        for score in scores:
            if score[0] in self.subset:
                sub_results.append(score)
        return sub_results


def _ngram_export(data):
    if isinstance(data, dict):
        to_implode = []
        for key, value in data.items():
            to_implode.append(repr(key) + ':' + _ngram_export(value))
        code = '{' + ','.join(to_implode) + '}'
        return code
    else:
        return repr(data)


def _save_ngrams(file_path, lang_array):
    if not os.path.exists(file_path):  # in case self.loaded_subset != new_subset, and was previously saved
        try:
            with open(file_path, 'w') as f:
                f.write(
                    '# Copyright 2023 Nito T.M. [ Apache 2.0 Licence https://www.apache.org/licenses/LICENSE-2.0 ]\n' +
                    'ngrams_data = {\n' +
                    '   "type": "' + str(languageData.type) + '",\n' +
                    '   "languages": ' + str(_iso_languages(lang_array)) + ',\n' +
                    '   "is_subset": True,\n' +
                    '   "ngrams": ' + _ngram_export(languageData.ngrams) + '\n' +
                    '}')
        except Exception as e:
            logging.exception(e)
            return False
    return True


def _make_subset(languages):
    subset = []
    reverse_langs = {v: k for k, v in languageData.lang_codes.items()}
    if languages:
        for lang in languages:
            found_lang = reverse_langs.get(lang)
            if found_lang is not None:
                subset.append(found_lang)
        subset.sort()
    return subset or None


def _iso_languages(lang_set):
    """Converts ngram database language indexes (integer) to ISO 639-1 code"""
    lang_codes = {}
    for lang_id in lang_set:
        lang_codes[lang_id] = languageData.lang_codes[lang_id]
    return lang_codes


def base16_to_base36(hex_string):
    # Convert hex string to integer
    integer_value = int(hex_string, 16)
    # Convert integer to base-36 string
    base36_string = ''
    while integer_value > 0:
        integer_value, remainder = divmod(integer_value, 36)
        base36_digit = '0123456789abcdefghijklmnopqrstuvwxyz'[remainder]
        base36_string = base36_digit + base36_string

    return base36_string



# Copyright 2023 Nito T.M.
# License https://www.apache.org/licenses/LICENSE-2.0 Apache-2.0
# Author Nito T.M. (https://github.com/nitotm)
# Package pypi.org/project/eld/

class SubsetResult:
    def __init__(self, success, languages=None, error=None, file=None):
        self.success = success
        self.languages = list(languages.values()) if languages else None
        self.error = error
        self.file = file



import pytest
import os
import sys

# Make sure, local package is imported instead of pip package
project_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))
sys.path.insert(0, project_root)  # prioritize the local package
# sys.path.append('../..')

from eld import LanguageDetector
from eld.languageDetector import get_clean_txt


# Mostly functional testing, when functions are more mature I will add some more unit tests

def test_print_version():
    detector = LanguageDetector()
    print('ELD ver. ' + detector.VERSION)
    assert True


def test_load_eld():
    detector = LanguageDetector()
    assert isinstance(detector, LanguageDetector)


def test_simple_detect():
    detector = LanguageDetector()
    result = detector.detect('Hola, cómo te llamas?').language
    assert result == 'es'


def test_get_multiple_scores():
    detector = LanguageDetector()
    detector.return_scores = True
    result = len(detector.detect('Hola, cómo te llamas?').scores())
    assert result > 1, 'Expected: >1 scores'


def test_detect_error_empty_text():
    detector = LanguageDetector()
    result = detector.detect('').language
    assert result is None


def test_clean_text():
    text = 'https://www.google.com/\n' \
           'mail@gmail.com\n' \
           'oogle.com/search?q=search&source=hp\n' \
           '12345 A12345\n'
    result = get_clean_txt(text).strip()
    assert result == ''


def test_check_confidence():
    detector = LanguageDetector('ngramsM60')
    text = 'zxz zcz zvz zbz znz zmz zlz zsz zdz zkz zjz pelo'
    result = detector.detect(text).is_reliable()
    assert result is False


def test_load_ngrams_detect():
    detector = LanguageDetector('ngramsM60-6_5ijqhj4oecs310zqtm8u9pgmd9ox2yd')
    result = detector.detect('Hola, cómo te llamas?').language
    assert result == 'es'


def test_accuracy_m_bigtest():
    # TODO use importlib or pathlib to open txt file as package eld.tests.data resource
    detector = LanguageDetector('ngramsM60')
    file = open( os.path.dirname(__file__) + '/data/big-test.txt', encoding='utf-8')
    # '../../benchmark/big-test.txt'
    content = file.read()
    file.close()
    lines = content.strip().split("\n")
    total = 0
    correct = 0
    for line in lines:
        total += 1
        values = line.split("\t")
        if detector.detect(values[1]).language == values[0]:
            correct += 1
    if total < 60000:
        pytest.fail('big-test.txt was not load correctly, too few lines')
    result = correct / total * 100
    # a bit of margin, depending on tie scores order, avg. might change a bit
    assert result > 99.4

# python -m pytest -v -s test_detector.py
# if __name__ == '__main__':
#    pytest.main(["-v", "test_detector.py"])  # Gives errors



# Copyright 2023 Nito T.M.
# License https://www.apache.org/licenses/LICENSE-2.0 Apache-2.0
# Author Nito T.M. (https://github.com/nitotm)
# Package pypi.org/project/eld/

import regex as re

from eld.languageData import languageData
from eld.languageSubset import LanguageSubset
from eld.languageResult import LanguageResult


class LanguageDetector(LanguageSubset):
    def __init__(self, subset_file=''):
        super().__init__()
        languageData.load_ngrams(subset_file)
        self.__do_clean_text = False
        self.VERSION = '1.0.6'  # Has to match setup.py version

    def detect(self, text):
        """
        Returns the language detected for a given UTF-8 string, as an ISO 639-1 code
            LanguageResult object { language = 'es', scores() = {'es': 0.5, 'et': 0.2}, is_reliable() = True }

        Args:
            text (str): UTF-8 string

        Returns:
            object LanguageResult: language (str or None), scores() (dict or None), is_reliable() (bool)
        """
        if self.__do_clean_text:
            # Removes Urls, emails, alphanumerical & numbers
            text = get_clean_txt(text)
        text = _normalize_text(text)
        txt_ngrams = _get_byte_ngrams(text)
        num_ngrams = len(txt_ngrams)
        results = _calculate_scores(txt_ngrams, num_ngrams)

        if results:
            if self.subset:
                results = LanguageSubset._filter_lang_subset(self, results)
            results.sort(key=lambda x: -x[1])
            return LanguageResult(results, num_ngrams)
        return LanguageResult()

    def clean_text(self, set_bool):
        self.__do_clean_text = (True if set_bool else False)


def _tokenizer(txt):
    return filter(None, re.split(b'\x20', txt))


def get_clean_txt(txt):
    """Removes parts of a string, that may be considered as "noise" for language detection"""
    # Remove URLS
    txt = re.sub(r'[hw]((ttps?://(www\.)?)|ww\.)([^\s/?\.#-]+\.?)+(\/\S*)?', ' ', txt, flags=re.IGNORECASE)
    # Remove emails
    txt = re.sub(r'[a-zA-Z0-9.!$%&?+_`-]+@[A-Za-z0-9.-]+\.[A-Za-z0-9-]{2,64}', ' ', txt)
    # Remove .com domains
    txt = re.sub(r'([A-Za-z0-9-]+\.)+com(\/\S*|[^\pL])', ' ', txt)
    # Remove alphanumerical/number codes
    txt = re.sub(r'[a-zA-Z]*[0-9]+[a-zA-Z0-9]*', ' ', txt)
    return txt


def _normalize_text(text):
    """Normalize special characters/word separators"""
    text = re.sub(r'[^\pL]+(?<![\x27\x60\x2019])', ' ', text[:1000], flags=re.UNICODE).strip()
    text = text.lower()
    text = bytes(text, 'utf-8')
    this_length = len(text)

    if this_length > 350:
        # Cut to first whitespace after 350 byte length offset
        text = text[0:min(380, (text.find(b'\x20', 350) or 350))]
    return text


def _calculate_scores(txt_ngrams, num_ngrams):
    """Calculate scores for each language from the given Ngrams"""
    lang_score = languageData.lang_score[:]
    for bytes_, frequency in txt_ngrams.items():
        if bytes_ in languageData.ngrams:
            lang_count = len(languageData.ngrams[bytes_])
            # Ngram score multiplier, the fewer languages found the more relevancy. Formula can be fine-tuned.
            if lang_count == 1:
                relevancy = 27
            elif lang_count < 16:
                relevancy = (16 - lang_count) / 2 + 1
            else:
                relevancy = 1

            # Most time-consuming loop, do only the strictly necessary inside
            for lang, global_frequency in languageData.ngrams[bytes_].items():
                lang_score[lang] += (global_frequency / frequency if frequency > global_frequency
                                     else frequency / global_frequency) * relevancy + 2
    # This divisor will produce a final score between 0 - ~1, score could be >1. Can be improved.
    result_divisor = num_ngrams * 3.2
    results = []
    for lang in range(len(lang_score)):
        if lang_score[lang]:
            results.append([lang, lang_score[lang] / result_divisor])  # * languageData.scoreNormalizer[lang]
    return results


def _get_byte_ngrams(txt):
    """Gets Ngrams from a given string"""
    byte_grams = {}
    count_ngrams = 0

    for word in _tokenizer(txt):
        length = len(word)

        if length > 70:
            length = 70
        x = 0
        for j in range(0, length - 4, 3):
            this_bytes = (b' ' if j == 0 else b'') + word[j:j + 4]
            byte_grams[this_bytes] = (1 + byte_grams[this_bytes] if this_bytes in byte_grams else 1)
            count_ngrams += 1
            x = 1

        this_bytes = (b' ' if x == 0 else b'') + word[length - 4 if length != 3 else 0:] + b' '
        byte_grams[this_bytes] = (1 + byte_grams[this_bytes] if this_bytes in byte_grams else 1)
        count_ngrams += 1

    # Frequency is multiplied by 15000 at the ngrams database. A reduced number seems to work better.
    # Linear formulas were tried, decreasing the multiplier for fewer ngram strings, no meaningful improvement.
    for bytes_, count in byte_grams.items():
        byte_grams[bytes_] = count / count_ngrams * 13200

    return byte_grams
